'''
Created on 06-Jul-2018

@author: abhinav
'''

LOG_TYPE_SERVER_INFO = 1000
LOG_TYPE_PROCESSED_SQS_MESSAGE = 1001
LOG_TYPE_EXCEPTION = -1
LOG_TYPE_DEBUG = 2


TYPE_OK = 1
TYPE_ERROR = -1
TYPE_NEED_LOGIN = -2